@@ -5,6 +5,7 @@
 use crate::buf::CffiBuf;
 use crate::error::{CryptographyError, CryptographyResult};
 use crate::{exceptions, types};
+use pyo3::types::{PyAnyMethods, PyListMethods};
 
 fn check_length(data: &[u8]) -> CryptographyResult<()> {
     if data.len() > (i32::MAX as usize) {
@@ -21,7 +22,7 @@ fn check_length(data: &[u8]) -> CryptographyResult<()> {
 
 enum Aad<'a> {
     Single(CffiBuf<'a>),
-    List(&'a pyo3::types::PyList),
+    List(pyo3::Bound<'a, pyo3::types::PyList>),
 }
 
 struct EvpCipherAead {
@@ -77,6 +78,7 @@ impl EvpCipherAead {
         ctx: &mut openssl::cipher_ctx::CipherCtx,
         data: &[u8],
         out: &mut [u8],
+        is_ccm: bool,
     ) -> CryptographyResult<()> {
         let bs = ctx.block_size();
 
@@ -87,9 +89,11 @@ impl EvpCipherAead {
             let n = ctx.cipher_update(data, Some(out))?;
             assert_eq!(n, data.len());
 
-            let mut final_block = [0];
-            let n = ctx.cipher_final(&mut final_block)?;
-            assert_eq!(n, 0);
+            if !is_ccm {
+                let mut final_block = [0];
+                let n = ctx.cipher_final(&mut final_block)?;
+                assert_eq!(n, 0);
+            }
         } else {
             // Our algorithm here is: split the data into the full chunks, and
             // the remaining partial chunk. Feed the full chunks into OpenSSL
@@ -128,12 +132,22 @@ impl EvpCipherAead {
         plaintext: &[u8],
         aad: Option<Aad<'_>>,
         nonce: Option<&[u8]>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let mut ctx = openssl::cipher_ctx::CipherCtx::new()?;
         ctx.copy(&self.base_encryption_ctx)?;
-        Self::encrypt_with_context(py, ctx, plaintext, aad, nonce, self.tag_len, self.tag_first)
+        Self::encrypt_with_context(
+            py,
+            ctx,
+            plaintext,
+            aad,
+            nonce,
+            self.tag_len,
+            self.tag_first,
+            false,
+        )
     }
 
+    #[allow(clippy::too_many_arguments)]
     fn encrypt_with_context<'p>(
         py: pyo3::Python<'p>,
         mut ctx: openssl::cipher_ctx::CipherCtx,
@@ -142,13 +156,19 @@ impl EvpCipherAead {
         nonce: Option<&[u8]>,
         tag_len: usize,
         tag_first: bool,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+        is_ccm: bool,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         check_length(plaintext)?;
 
-        if let Some(nonce) = nonce {
-            ctx.set_iv_length(nonce.len())?;
+        if !is_ccm {
+            if let Some(nonce) = nonce {
+                ctx.set_iv_length(nonce.len())?;
+            }
+            ctx.encrypt_init(None, None, nonce)?;
+        }
+        if is_ccm {
+            ctx.set_data_len(plaintext.len())?;
         }
-        ctx.encrypt_init(None, None, nonce)?;
 
         Self::process_aad(&mut ctx, aad)?;
 
@@ -164,7 +184,7 @@ impl EvpCipherAead {
                     (ciphertext, tag) = b.split_at_mut(plaintext.len());
                 }
 
-                Self::process_data(&mut ctx, plaintext, ciphertext)?;
+                Self::process_data(&mut ctx, plaintext, ciphertext, is_ccm)?;
 
                 ctx.tag(tag).map_err(CryptographyError::from)?;
 
@@ -179,7 +199,7 @@ impl EvpCipherAead {
         ciphertext: &[u8],
         aad: Option<Aad<'_>>,
         nonce: Option<&[u8]>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let mut ctx = openssl::cipher_ctx::CipherCtx::new()?;
         ctx.copy(&self.base_decryption_ctx)?;
         Self::decrypt_with_context(
@@ -190,9 +210,11 @@ impl EvpCipherAead {
             nonce,
             self.tag_len,
             self.tag_first,
+            false,
         )
     }
 
+    #[allow(clippy::too_many_arguments)]
     fn decrypt_with_context<'p>(
         py: pyo3::Python<'p>,
         mut ctx: openssl::cipher_ctx::CipherCtx,
@@ -201,16 +223,12 @@ impl EvpCipherAead {
         nonce: Option<&[u8]>,
         tag_len: usize,
         tag_first: bool,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+        is_ccm: bool,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         if ciphertext.len() < tag_len {
             return Err(CryptographyError::from(exceptions::InvalidTag::new_err(())));
         }
 
-        if let Some(nonce) = nonce {
-            ctx.set_iv_length(nonce.len())?;
-        }
-        ctx.decrypt_init(None, None, nonce)?;
-
         let tag;
         let ciphertext_data;
         if tag_first {
@@ -221,7 +239,18 @@ impl EvpCipherAead {
         } else {
             (ciphertext_data, tag) = ciphertext.split_at(ciphertext.len() - tag_len);
         }
-        ctx.set_tag(tag)?;
+
+        if !is_ccm {
+            if let Some(nonce) = nonce {
+                ctx.set_iv_length(nonce.len())?;
+            }
+
+            ctx.decrypt_init(None, None, nonce)?;
+            ctx.set_tag(tag)?;
+        }
+        if is_ccm {
+            ctx.set_data_len(ciphertext_data.len())?;
+        }
 
         Self::process_aad(&mut ctx, aad)?;
 
@@ -229,7 +258,7 @@ impl EvpCipherAead {
             py,
             ciphertext_data.len(),
             |b| {
-                Self::process_data(&mut ctx, ciphertext_data, b)
+                Self::process_data(&mut ctx, ciphertext_data, b, is_ccm)
                     .map_err(|_| exceptions::InvalidTag::new_err(()))?;
 
                 Ok(())
@@ -238,38 +267,30 @@ impl EvpCipherAead {
     }
 }
 
-#[cfg(not(any(
-    CRYPTOGRAPHY_IS_LIBRESSL,
-    CRYPTOGRAPHY_IS_BORINGSSL,
-    not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER),
-    CRYPTOGRAPHY_OPENSSL_320_OR_GREATER
-)))]
 struct LazyEvpCipherAead {
     cipher: &'static openssl::cipher::CipherRef,
     key: pyo3::Py<pyo3::PyAny>,
 
     tag_len: usize,
     tag_first: bool,
+    is_ccm: bool,
 }
 
-#[cfg(not(any(
-    CRYPTOGRAPHY_IS_LIBRESSL,
-    CRYPTOGRAPHY_IS_BORINGSSL,
-    not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER),
-    CRYPTOGRAPHY_OPENSSL_320_OR_GREATER
-)))]
 impl LazyEvpCipherAead {
+    #[cfg(not(CRYPTOGRAPHY_IS_BORINGSSL))]
     fn new(
         cipher: &'static openssl::cipher::CipherRef,
         key: pyo3::Py<pyo3::PyAny>,
         tag_len: usize,
         tag_first: bool,
+        is_ccm: bool,
     ) -> LazyEvpCipherAead {
         LazyEvpCipherAead {
             cipher,
             key,
             tag_len,
             tag_first,
+            is_ccm,
         }
     }
 
@@ -279,11 +300,19 @@ impl LazyEvpCipherAead {
         plaintext: &[u8],
         aad: Option<Aad<'_>>,
         nonce: Option<&[u8]>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
-        let key_buf = self.key.as_ref(py).extract::<CffiBuf<'_>>()?;
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        let key_buf = self.key.bind(py).extract::<CffiBuf<'_>>()?;
 
         let mut encryption_ctx = openssl::cipher_ctx::CipherCtx::new()?;
-        encryption_ctx.encrypt_init(Some(self.cipher), Some(key_buf.as_bytes()), None)?;
+        if self.is_ccm {
+            encryption_ctx.encrypt_init(Some(self.cipher), None, None)?;
+            encryption_ctx.set_iv_length(nonce.as_ref().unwrap().len())?;
+            encryption_ctx.set_tag_length(self.tag_len)?;
+            encryption_ctx.encrypt_init(None, Some(key_buf.as_bytes()), nonce)?;
+        } else {
+            encryption_ctx.encrypt_init(Some(self.cipher), Some(key_buf.as_bytes()), None)?;
+        }
+
         EvpCipherAead::encrypt_with_context(
             py,
             encryption_ctx,
@@ -292,6 +321,7 @@ impl LazyEvpCipherAead {
             nonce,
             self.tag_len,
             self.tag_first,
+            self.is_ccm,
         )
     }
 
@@ -301,11 +331,26 @@ impl LazyEvpCipherAead {
         ciphertext: &[u8],
         aad: Option<Aad<'_>>,
         nonce: Option<&[u8]>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
-        let key_buf = self.key.as_ref(py).extract::<CffiBuf<'_>>()?;
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        let key_buf = self.key.bind(py).extract::<CffiBuf<'_>>()?;
 
         let mut decryption_ctx = openssl::cipher_ctx::CipherCtx::new()?;
-        decryption_ctx.decrypt_init(Some(self.cipher), Some(key_buf.as_bytes()), None)?;
+        if self.is_ccm {
+            decryption_ctx.decrypt_init(Some(self.cipher), None, None)?;
+            decryption_ctx.set_iv_length(nonce.as_ref().unwrap().len())?;
+
+            if ciphertext.len() < self.tag_len {
+                return Err(CryptographyError::from(exceptions::InvalidTag::new_err(())));
+            }
+
+            let (_, tag) = ciphertext.split_at(ciphertext.len() - self.tag_len);
+            decryption_ctx.set_tag(tag)?;
+
+            decryption_ctx.decrypt_init(None, Some(key_buf.as_bytes()), nonce)?;
+        } else {
+            decryption_ctx.decrypt_init(Some(self.cipher), Some(key_buf.as_bytes()), None)?;
+        }
+
         EvpCipherAead::decrypt_with_context(
             py,
             decryption_ctx,
@@ -314,6 +359,7 @@ impl LazyEvpCipherAead {
             nonce,
             self.tag_len,
             self.tag_first,
+            self.is_ccm,
         )
     }
 }
@@ -343,7 +389,7 @@ impl EvpAead {
         plaintext: &[u8],
         aad: Option<Aad<'_>>,
         nonce: Option<&[u8]>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         check_length(plaintext)?;
 
         let ad = if let Some(Aad::Single(ad)) = &aad {
@@ -371,7 +417,7 @@ impl EvpAead {
         ciphertext: &[u8],
         aad: Option<Aad<'_>>,
         nonce: Option<&[u8]>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         if ciphertext.len() < self.tag_len {
             return Err(CryptographyError::from(exceptions::InvalidTag::new_err(())));
         }
@@ -398,7 +444,7 @@ impl EvpAead {
     }
 }
 
-#[pyo3::prelude::pyclass(frozen, module = "cryptography.hazmat.bindings._rust.openssl.aead")]
+#[pyo3::pyclass(frozen, module = "cryptography.hazmat.bindings._rust.openssl.aead")]
 struct ChaCha20Poly1305 {
     #[cfg(CRYPTOGRAPHY_IS_BORINGSSL)]
     ctx: EvpAead,
@@ -420,7 +466,7 @@ struct ChaCha20Poly1305 {
     ctx: LazyEvpCipherAead,
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl ChaCha20Poly1305 {
     #[new]
     fn new(py: pyo3::Python<'_>, key: pyo3::Py<pyo3::PyAny>) -> CryptographyResult<Self> {
@@ -443,8 +489,8 @@ impl ChaCha20Poly1305 {
             } else if #[cfg(any(
                 CRYPTOGRAPHY_IS_LIBRESSL,
                 CRYPTOGRAPHY_OPENSSL_320_OR_GREATER,
-                not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER
-            )))] {
+                not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER),
+            ))] {
                 if cryptography_openssl::fips::is_enabled() {
                     return Err(CryptographyError::from(
                         exceptions::UnsupportedAlgorithm::new_err((
@@ -478,6 +524,7 @@ impl ChaCha20Poly1305 {
                         key,
                         16,
                         false,
+                        false,
                     )
                 })
             }
@@ -485,19 +532,18 @@ impl ChaCha20Poly1305 {
     }
 
     #[staticmethod]
-    fn generate_key(py: pyo3::Python<'_>) -> CryptographyResult<&pyo3::PyAny> {
-        Ok(py
-            .import(pyo3::intern!(py, "os"))?
-            .call_method1(pyo3::intern!(py, "urandom"), (32,))?)
+    fn generate_key(py: pyo3::Python<'_>) -> CryptographyResult<pyo3::Bound<'_, pyo3::PyAny>> {
+        Ok(types::OS_URANDOM.get(py)?.call1((32,))?)
     }
 
+    #[pyo3(signature = (nonce, data, associated_data))]
     fn encrypt<'p>(
         &self,
         py: pyo3::Python<'p>,
         nonce: CffiBuf<'_>,
         data: CffiBuf<'_>,
         associated_data: Option<CffiBuf<'_>>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let nonce_bytes = nonce.as_bytes();
         let aad = associated_data.map(Aad::Single);
 
@@ -511,13 +557,14 @@ impl ChaCha20Poly1305 {
             .encrypt(py, data.as_bytes(), aad, Some(nonce_bytes))
     }
 
+    #[pyo3(signature = (nonce, data, associated_data))]
     fn decrypt<'p>(
         &self,
         py: pyo3::Python<'p>,
         nonce: CffiBuf<'_>,
         data: CffiBuf<'_>,
         associated_data: Option<CffiBuf<'_>>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let nonce_bytes = nonce.as_bytes();
         let aad = associated_data.map(Aad::Single);
 
@@ -532,7 +579,265 @@ impl ChaCha20Poly1305 {
     }
 }
 
-#[pyo3::prelude::pyclass(
+#[pyo3::pyclass(
+    frozen,
+    module = "cryptography.hazmat.bindings._rust.openssl.aead",
+    name = "AESGCM"
+)]
+struct AesGcm {
+    #[cfg(any(
+        CRYPTOGRAPHY_OPENSSL_320_OR_GREATER,
+        CRYPTOGRAPHY_IS_LIBRESSL,
+        CRYPTOGRAPHY_IS_BORINGSSL,
+        not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER),
+    ))]
+    ctx: EvpCipherAead,
+
+    #[cfg(not(any(
+        CRYPTOGRAPHY_OPENSSL_320_OR_GREATER,
+        CRYPTOGRAPHY_IS_LIBRESSL,
+        CRYPTOGRAPHY_IS_BORINGSSL,
+        not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER),
+    )))]
+    ctx: LazyEvpCipherAead,
+}
+
+#[pyo3::pymethods]
+impl AesGcm {
+    #[new]
+    fn new(py: pyo3::Python<'_>, key: pyo3::Py<pyo3::PyAny>) -> CryptographyResult<AesGcm> {
+        let key_buf = key.extract::<CffiBuf<'_>>(py)?;
+        let cipher = match key_buf.as_bytes().len() {
+            16 => openssl::cipher::Cipher::aes_128_gcm(),
+            24 => openssl::cipher::Cipher::aes_192_gcm(),
+            32 => openssl::cipher::Cipher::aes_256_gcm(),
+            _ => {
+                return Err(CryptographyError::from(
+                    pyo3::exceptions::PyValueError::new_err(
+                        "AESGCM key must be 128, 192, or 256 bits.",
+                    ),
+                ))
+            }
+        };
+
+        cfg_if::cfg_if! {
+            if #[cfg(any(
+                CRYPTOGRAPHY_OPENSSL_320_OR_GREATER,
+                CRYPTOGRAPHY_IS_BORINGSSL,
+                CRYPTOGRAPHY_IS_LIBRESSL,
+                not(CRYPTOGRAPHY_OPENSSL_300_OR_GREATER),
+            ))] {
+                Ok(AesGcm {
+                    ctx: EvpCipherAead::new(cipher, key_buf.as_bytes(), 16, false)?,
+                })
+            } else {
+                Ok(AesGcm {
+                    ctx: LazyEvpCipherAead::new(cipher, key, 16, false, false),
+                })
+
+            }
+        }
+    }
+
+    #[staticmethod]
+    fn generate_key(
+        py: pyo3::Python<'_>,
+        bit_length: usize,
+    ) -> CryptographyResult<pyo3::Bound<'_, pyo3::PyAny>> {
+        if bit_length != 128 && bit_length != 192 && bit_length != 256 {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("bit_length must be 128, 192, or 256"),
+            ));
+        }
+
+        Ok(types::OS_URANDOM.get(py)?.call1((bit_length / 8,))?)
+    }
+
+    #[pyo3(signature = (nonce, data, associated_data))]
+    fn encrypt<'p>(
+        &self,
+        py: pyo3::Python<'p>,
+        nonce: CffiBuf<'_>,
+        data: CffiBuf<'_>,
+        associated_data: Option<CffiBuf<'_>>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        let nonce_bytes = nonce.as_bytes();
+        let aad = associated_data.map(Aad::Single);
+
+        if nonce_bytes.len() < 8 || nonce_bytes.len() > 128 {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("Nonce must be between 8 and 128 bytes"),
+            ));
+        }
+
+        self.ctx
+            .encrypt(py, data.as_bytes(), aad, Some(nonce_bytes))
+    }
+
+    #[pyo3(signature = (nonce, data, associated_data))]
+    fn decrypt<'p>(
+        &self,
+        py: pyo3::Python<'p>,
+        nonce: CffiBuf<'_>,
+        data: CffiBuf<'_>,
+        associated_data: Option<CffiBuf<'_>>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        let nonce_bytes = nonce.as_bytes();
+        let aad = associated_data.map(Aad::Single);
+
+        if nonce_bytes.len() < 8 || nonce_bytes.len() > 128 {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("Nonce must be between 8 and 128 bytes"),
+            ));
+        }
+
+        self.ctx
+            .decrypt(py, data.as_bytes(), aad, Some(nonce_bytes))
+    }
+}
+
+#[pyo3::pyclass(
+    frozen,
+    module = "cryptography.hazmat.bindings._rust.openssl.aead",
+    name = "AESCCM"
+)]
+struct AesCcm {
+    ctx: LazyEvpCipherAead,
+    tag_length: usize,
+}
+
+#[pyo3::pymethods]
+impl AesCcm {
+    #[new]
+    #[pyo3(signature = (key, tag_length=None))]
+    fn new(
+        py: pyo3::Python<'_>,
+        key: pyo3::Py<pyo3::PyAny>,
+        tag_length: Option<usize>,
+    ) -> CryptographyResult<AesCcm> {
+        cfg_if::cfg_if! {
+            if #[cfg(CRYPTOGRAPHY_IS_BORINGSSL)] {
+                let _ = py;
+                let _ = key;
+                let _ = tag_length;
+                Err(CryptographyError::from(
+                    exceptions::UnsupportedAlgorithm::new_err((
+                        "AES-CCM is not supported by this version of OpenSSL",
+                        exceptions::Reasons::UNSUPPORTED_CIPHER,
+                    )),
+                ))
+            } else {
+                let key_buf = key.extract::<CffiBuf<'_>>(py)?;
+                let cipher = match key_buf.as_bytes().len() {
+                    16 => openssl::cipher::Cipher::aes_128_ccm(),
+                    24 => openssl::cipher::Cipher::aes_192_ccm(),
+                    32 => openssl::cipher::Cipher::aes_256_ccm(),
+                    _ => {
+                        return Err(CryptographyError::from(
+                            pyo3::exceptions::PyValueError::new_err(
+                                "AESCCM key must be 128, 192, or 256 bits.",
+                            ),
+                        ))
+                    }
+                };
+                let tag_length = tag_length.unwrap_or(16);
+                if ![4, 6, 8, 10, 12, 14, 16].contains(&tag_length) {
+                    return Err(CryptographyError::from(
+                        pyo3::exceptions::PyValueError::new_err("Invalid tag_length"),
+                    ));
+                }
+
+                Ok(AesCcm {
+                    ctx: LazyEvpCipherAead::new(cipher, key, tag_length, false, true),
+                    tag_length
+                })
+            }
+        }
+    }
+
+    #[staticmethod]
+    fn generate_key(
+        py: pyo3::Python<'_>,
+        bit_length: usize,
+    ) -> CryptographyResult<pyo3::Bound<'_, pyo3::PyAny>> {
+        if bit_length != 128 && bit_length != 192 && bit_length != 256 {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("bit_length must be 128, 192, or 256"),
+            ));
+        }
+
+        Ok(types::OS_URANDOM.get(py)?.call1((bit_length / 8,))?)
+    }
+
+    #[pyo3(signature = (nonce, data, associated_data))]
+    fn encrypt<'p>(
+        &self,
+        py: pyo3::Python<'p>,
+        nonce: CffiBuf<'_>,
+        data: CffiBuf<'_>,
+        associated_data: Option<CffiBuf<'_>>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        let nonce_bytes = nonce.as_bytes();
+        let data_bytes = data.as_bytes();
+        let aad = associated_data.map(Aad::Single);
+
+        if nonce_bytes.len() < 7 || nonce_bytes.len() > 13 {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("Nonce must be between 7 and 13 bytes"),
+            ));
+        }
+
+        check_length(data_bytes)?;
+        // For information about computing this, see
+        // https://tools.ietf.org/html/rfc3610#section-2.1
+        let l_val = 15 - nonce_bytes.len();
+        let max_length = 1usize.checked_shl(8 * l_val as u32);
+        // If `max_length` overflowed, then it's not possible for data to be
+        // longer than it.
+        if max_length.map(|v| v < data_bytes.len()).unwrap_or(false) {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("Data too long for nonce"),
+            ));
+        }
+
+        self.ctx.encrypt(py, data_bytes, aad, Some(nonce_bytes))
+    }
+
+    #[pyo3(signature = (nonce, data, associated_data))]
+    fn decrypt<'p>(
+        &self,
+        py: pyo3::Python<'p>,
+        nonce: CffiBuf<'_>,
+        data: CffiBuf<'_>,
+        associated_data: Option<CffiBuf<'_>>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
+        let nonce_bytes = nonce.as_bytes();
+        let data_bytes = data.as_bytes();
+        let aad = associated_data.map(Aad::Single);
+
+        if nonce_bytes.len() < 7 || nonce_bytes.len() > 13 {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("Nonce must be between 7 and 13 bytes"),
+            ));
+        }
+        // For information about computing this, see
+        // https://tools.ietf.org/html/rfc3610#section-2.1
+        let l_val = 15 - nonce_bytes.len();
+        let max_length = 1usize.checked_shl(8 * l_val as u32);
+        // If `max_length` overflowed, then it's not possible for data to be
+        // longer than it.
+        let pt_length = data_bytes.len().saturating_sub(self.tag_length);
+        if max_length.map(|v| v < pt_length).unwrap_or(false) {
+            return Err(CryptographyError::from(
+                pyo3::exceptions::PyValueError::new_err("Data too long for nonce"),
+            ));
+        }
+
+        self.ctx.decrypt(py, data_bytes, aad, Some(nonce_bytes))
+    }
+}
+
+#[pyo3::pyclass(
     frozen,
     module = "cryptography.hazmat.bindings._rust.openssl.aead",
     name = "AESSIV"
@@ -541,7 +846,7 @@ struct AesSiv {
     ctx: EvpCipherAead,
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl AesSiv {
     #[new]
     fn new(key: CffiBuf<'_>) -> CryptographyResult<AesSiv> {
@@ -574,19 +879,23 @@ impl AesSiv {
                     ctx: EvpCipherAead::new(&cipher, key.as_bytes(), 16, true)?,
                 })
             } else {
-                return Err(CryptographyError::from(
+                _ = cipher_name;
+
+                Err(CryptographyError::from(
                     exceptions::UnsupportedAlgorithm::new_err((
                         "AES-SIV is not supported by this version of OpenSSL",
                         exceptions::Reasons::UNSUPPORTED_CIPHER,
                     )),
-                ));
-
+                ))
             }
         }
     }
 
     #[staticmethod]
-    fn generate_key(py: pyo3::Python<'_>, bit_length: usize) -> CryptographyResult<&pyo3::PyAny> {
+    fn generate_key(
+        py: pyo3::Python<'_>,
+        bit_length: usize,
+    ) -> CryptographyResult<pyo3::Bound<'_, pyo3::PyAny>> {
         if bit_length != 256 && bit_length != 384 && bit_length != 512 {
             return Err(CryptographyError::from(
                 pyo3::exceptions::PyValueError::new_err("bit_length must be 256, 384, or 512"),
@@ -601,8 +910,8 @@ impl AesSiv {
         &self,
         py: pyo3::Python<'p>,
         data: CffiBuf<'_>,
-        associated_data: Option<&pyo3::types::PyList>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+        associated_data: Option<pyo3::Bound<'p, pyo3::types::PyList>>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let data_bytes = data.as_bytes();
         let aad = associated_data.map(Aad::List);
 
@@ -619,14 +928,14 @@ impl AesSiv {
         &self,
         py: pyo3::Python<'p>,
         data: CffiBuf<'_>,
-        associated_data: Option<&pyo3::types::PyList>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+        associated_data: Option<pyo3::Bound<'_, pyo3::types::PyList>>,
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let aad = associated_data.map(Aad::List);
         self.ctx.decrypt(py, data.as_bytes(), aad, None)
     }
 }
 
-#[pyo3::prelude::pyclass(
+#[pyo3::pyclass(
     frozen,
     module = "cryptography.hazmat.bindings._rust.openssl.aead",
     name = "AESOCB3"
@@ -635,18 +944,20 @@ struct AesOcb3 {
     ctx: EvpCipherAead,
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl AesOcb3 {
     #[new]
     fn new(key: CffiBuf<'_>) -> CryptographyResult<AesOcb3> {
         cfg_if::cfg_if! {
             if #[cfg(any(CRYPTOGRAPHY_IS_LIBRESSL, CRYPTOGRAPHY_IS_BORINGSSL))] {
-                return Err(CryptographyError::from(
+                _ = key;
+
+                Err(CryptographyError::from(
                     exceptions::UnsupportedAlgorithm::new_err((
                         "AES-OCB3 is not supported by this version of OpenSSL",
                         exceptions::Reasons::UNSUPPORTED_CIPHER,
                     )),
-                ));
+                ))
             } else {
                 if cryptography_openssl::fips::is_enabled() {
                     return Err(CryptographyError::from(
@@ -678,7 +989,10 @@ impl AesOcb3 {
     }
 
     #[staticmethod]
-    fn generate_key(py: pyo3::Python<'_>, bit_length: usize) -> CryptographyResult<&pyo3::PyAny> {
+    fn generate_key(
+        py: pyo3::Python<'_>,
+        bit_length: usize,
+    ) -> CryptographyResult<pyo3::Bound<'_, pyo3::PyAny>> {
         if bit_length != 128 && bit_length != 192 && bit_length != 256 {
             return Err(CryptographyError::from(
                 pyo3::exceptions::PyValueError::new_err("bit_length must be 128, 192, or 256"),
@@ -695,7 +1009,7 @@ impl AesOcb3 {
         nonce: CffiBuf<'_>,
         data: CffiBuf<'_>,
         associated_data: Option<CffiBuf<'_>>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let nonce_bytes = nonce.as_bytes();
         let aad = associated_data.map(Aad::Single);
 
@@ -716,7 +1030,7 @@ impl AesOcb3 {
         nonce: CffiBuf<'_>,
         data: CffiBuf<'_>,
         associated_data: Option<CffiBuf<'_>>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let nonce_bytes = nonce.as_bytes();
         let aad = associated_data.map(Aad::Single);
 
@@ -731,7 +1045,7 @@ impl AesOcb3 {
     }
 }
 
-#[pyo3::prelude::pyclass(
+#[pyo3::pyclass(
     frozen,
     module = "cryptography.hazmat.bindings._rust.openssl.aead",
     name = "AESGCMSIV"
@@ -740,7 +1054,7 @@ struct AesGcmSiv {
     ctx: EvpCipherAead,
 }
 
-#[pyo3::prelude::pymethods]
+#[pyo3::pymethods]
 impl AesGcmSiv {
     #[new]
     fn new(key: CffiBuf<'_>) -> CryptographyResult<AesGcmSiv> {
@@ -784,7 +1098,10 @@ impl AesGcmSiv {
     }
 
     #[staticmethod]
-    fn generate_key(py: pyo3::Python<'_>, bit_length: usize) -> CryptographyResult<&pyo3::PyAny> {
+    fn generate_key(
+        py: pyo3::Python<'_>,
+        bit_length: usize,
+    ) -> CryptographyResult<pyo3::Bound<'_, pyo3::PyAny>> {
         if bit_length != 128 && bit_length != 192 && bit_length != 256 {
             return Err(CryptographyError::from(
                 pyo3::exceptions::PyValueError::new_err("bit_length must be 128, 192, or 256"),
@@ -801,7 +1118,7 @@ impl AesGcmSiv {
         nonce: CffiBuf<'_>,
         data: CffiBuf<'_>,
         associated_data: Option<CffiBuf<'_>>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let nonce_bytes = nonce.as_bytes();
         let data_bytes = data.as_bytes();
         let aad = associated_data.map(Aad::Single);
@@ -826,7 +1143,7 @@ impl AesGcmSiv {
         nonce: CffiBuf<'_>,
         data: CffiBuf<'_>,
         associated_data: Option<CffiBuf<'_>>,
-    ) -> CryptographyResult<&'p pyo3::types::PyBytes> {
+    ) -> CryptographyResult<pyo3::Bound<'p, pyo3::types::PyBytes>> {
         let nonce_bytes = nonce.as_bytes();
         let aad = associated_data.map(Aad::Single);
         if nonce_bytes.len() != 12 {
@@ -839,13 +1156,8 @@ impl AesGcmSiv {
     }
 }
 
-pub(crate) fn create_module(py: pyo3::Python<'_>) -> pyo3::PyResult<&pyo3::prelude::PyModule> {
-    let m = pyo3::prelude::PyModule::new(py, "aead")?;
-
-    m.add_class::<ChaCha20Poly1305>()?;
-    m.add_class::<AesSiv>()?;
-    m.add_class::<AesOcb3>()?;
-    m.add_class::<AesGcmSiv>()?;
-
-    Ok(m)
+#[pyo3::pymodule]
+pub(crate) mod aead {
+    #[pymodule_export]
+    use super::{AesCcm, AesGcm, AesGcmSiv, AesOcb3, AesSiv, ChaCha20Poly1305};
 }
